;MICRO-READERBOARD ORNAMENT
;DOT MATRIX VERSION FOR MICRO SIZED DISPLAYS

;LED dot matrix display: columns = cathode, rows = anode, e.g. Kingbright TC07-11
;ATTiny24 @ 1 MHz (internal R/C osc.)
;Fuses:  low = 0x42 (internal 8MHz, divide Clock by 8, BOD enabled)
;       high = 0x76 (BOD level 1.8V, preserve EEPROM through chip erase, disable RESET and SPI)

;PORTS:
;  PA0 = Row 1 (right)
;  PA1 = Row 2
;  PA2 = Column 3, RS232 Rx
;  PA3 = Row 4, RS232 GND
;  PA4 = Column 5 (bottom)
;  PA5 = Column 4
;  PA6 = Row 7 (left)
;  PA7 = Row 6
;  PB0 = Column 1 (top)
;  PB1 = Row 3
;  PB2 = Row 5
;  PB3 = Column 2

;VERSION HISTORY:
;  Version 0.8     2007-04-23  Arne Rossius
;  Version 0.8.1   2007-07-07   Arne Rossius
;    * fixed bug in charset (char 0x35 = '5')

.include "../../include/tn24def.inc"

.def	coldata = R9
.def	scrollspeed = R10
.def	char1 = R11
.def	char2 = R12
.def	char3 = R13
.def	char4 = R14
.def	char5 = R15

.def	temp = R16
.def	temp2 = R17
.def	col1 = R18
.def	col2 = R19
.def	col3 = R20
.def	col4 = R21
.def	col5 = R22
.def	delay = R23
.def	strpos = R24
.def	charlen = R25
;R26/27 used for buffer (RAM) pointer
;R28/29 used for string (EEPROM) pointer
;R30/31 used for charset (Flash) pointer

;===============================================================================

.dseg
	String:
		.byte 112
	RAMstack:
		.byte 16
		
.eseg
	.org 0x00
	.db 0xff, 8, "Hello World!   ", 0, 0

;===============================================================================

.cseg
.org 0x000
	rjmp	start
.org OVF0addr ;Timer 0 overflow interrupt
	rjmp	t0ovf
	
;===============================================================================
	
charset:
	.include "../charset.asm"
	
;===============================================================================

t0ovf:
	;Timer 0 overflow interrupt (~60 Hz)
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	;decrement delay counter
	dec	delay
	brne	t0ovf_end
	mov	delay, scrollspeed
	
	;shift display left by one column
	lsr	char1
	rol	col1
	lsr	char2
	rol	col2
	lsr	char3
	rol	col3
	lsr	char4
	rol	col4
	lsr	char5
	rol	col5
	dec	charlen
	cpi	charlen, 9 ;overflow -> next char
	brlo	t0ovf_end
nextchar:
	;load next character
	rcall	eeprom_read
	inc	YL
	mov	ZL, temp
	cpi	ZL, 0x00 ;end marker -> restart string
	breq	restart
	subi	ZL, 0x20 ;charset starts at 0x20
	clr	ZH
	lsl	ZL ;offset: multiply ZH:ZL by 6
	rol	ZH
	mov	temp, ZL
	mov	temp2, ZH
	lsl	ZL
	rol	ZH
	add	ZL, temp
	adc	ZH, temp2
	ldi	temp, LOW(charset*2) ;add charset address
	ldi	temp2, HIGH(charset*2)
	add	ZL, temp
	adc	ZH, temp2
	lpm	charlen, Z+ ;load and invert character
	lpm	char1, Z+
	lpm	char2, Z+
	lpm	char3, Z+
	lpm	char4, Z+
	lpm	char5, Z+
	com	char1
	com	char2
	com	char3
	com	char4
	com	char5
	
t0ovf_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti
	
restart:
	;restart string
	ldi	YL, 0x02
	rjmp	nextchar

;===============================================================================
	
start:
	;stack pointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;ports
	ldi	temp, 0xFB
	out	DDRA, temp
	ldi	temp, 0x0F
	out	DDRB, temp
	ldi	temp, 0x34
	out	PORTA, temp
	ldi	temp, 0x09
	out	PORTB, temp
	
	;check if RS232 cable is connected
	rcall	muxwait
	sbic	PINA, 2 ;PA2 low in spite of pullup -> RS232 connected
	rjmp	play

;===============================================================================

;---------------------------------------
;         RS232 TRANSFER MODE
;---------------------------------------

transfer:
	;init registers
	ldi	XL, LOW(String)
	ldi	XH, HIGH(String)
	ldi	YL, 0x02
	
getc:
	;wait for transmission to begin
	sbis	PINA, 2
	rjmp	getc
	
	;receive byte
	rcall	waitbit ;startbit
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 0
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit
	
	rcall	waitbit ;data 1
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 2
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 3
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 4
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 5
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 6
	clc
	sbis	PINA, 2
	sec
	ror	temp
	rcall	waitbit

	rcall	waitbit ;data 7
	clc
	sbis	PINA, 2
	sec
	ror	temp
	brcs	getc_starterr ;start bit was not 1
	rcall	waitbit

	rcall	waitbit
	sbic	PINA, 2
	rjmp	getc_stoperr ;stop bit is not 0
	
	;save byte to RAM buffer
	cpi	XL, String+112 ;prevent buffer overflow
	brlo	PC+2
	ldi	XL, String+111
	st	X+, temp

	;evaluate byte
	cpi	temp, 0x03 ;CTRL-C -> go to play mode
	breq	play
	cpi	temp, 0x04 ;CTRL-D -> evaluate string
	breq	getc_done
	cpi	temp, 0x0a ;LF -> evaluate string
	breq	getc_done
	cpi	temp, 0x0d ;CR -> evaluate string
	breq	getc_done
	cpi	temp, 0x08 ;BACKSP -> remove last character and the backspace character from buffer
	breq	getc_backsp
	rjmp	getc_end
	
getc_backsp:
	cpi	XL, String+2
	brlo	getc_end
	subi	XL, 2
	
getc_starterr:
getc_stoperr:
getc_end:
	;wait until line is idle (needed if stop bit is not 0) and receive next byte
	sbic	PINA, 2
	rjmp	getc_end
	rjmp	getc
	
;--------------------
	
getc_done:
	;terminate string
	dec	XL
	ldi	temp, 0x00 ;terminate string
	st	X, temp
getc_eval:
	;evaluate received string
	cpi	XL, String ;string is empty (contains only end marker)
	brne	PC+2
	rjmp	transfer
	ldi	XL, String
	ld	temp, X
	cpi	temp, 0x1b ;ESC
	breq	getc_control
	
	;save received string to EEPROM
	ldi	YL, 0x02
getc_savestr_loop:
	ld	temp, X+
	rcall	eeprom_write
	inc	YL
	cpi	temp, 0x00
	brne	getc_savestr_loop
	rjmp	transfer
	
getc_control:
	;control string received (begins with ESC)
	inc	XL
	ld	temp, X+
	cpi	temp, 's' ;scrolling speed
	breq	getc_speed
	cpi	temp, 'S'
	breq	getc_speed
	cpi	temp, 'q' ;quit -> switch to display mode
	breq	play
	cpi	temp, 'Q'
	breq	play
	cpi	temp, 0x1b ;ESC -> switch to display mode
	breq	play
	rjmp	transfer
	
getc_speed:
	;write scrolling speed setting to eeprom
	rcall	dec2bin
	ldi	YL, 0x01
	rcall	eeprom_write
	rjmp	transfer

;===============================================================================
	
;---------------------------------------
;             "PLAY" MODE
;---------------------------------------
	
play:
	;ports
	ldi	temp, 0xFF
	out	DDRA, temp
	ldi	temp, 0x34
	out	PORTA, temp
	ldi	temp, 0x0F
	out	DDRB, temp
	ldi	temp, 0x09
	out	PORTB, temp
	
	;scrolling speed
	ldi	YL, 0x01
	rcall	eeprom_read
	mov	scrollspeed, temp
	
	;initialize registers
	ldi	YL, 0x02
	clr	YH
	ldi	charlen, 0
	mov	delay, scrollspeed
	ldi	col1, 0xFF
	ldi	col2, 0xFF
	ldi	col3, 0xFF
	ldi	col4, 0xFF
	ldi	col5, 0xFF
	mov	char1, col1
	mov	char2, col2
	mov	char3, col3
	mov	char4, col4
	mov	char5, col5
	
	;timer 0
	ldi	temp, 0x03 ;Clk/64
	out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK0, temp
	
	;enable interrupts
	sei
	
	
loop:
	;multiplex display
	mov	coldata, col1
	rcall	setcol
	andi	temp, ~0x10
	out	PORTA, temp
	out	PORTB, temp2
	rcall	muxwait
	
	mov	coldata, col2
	rcall	setcol
	andi	temp, ~0x20
	out	PORTA, temp
	out	PORTB, temp2
	rcall	muxwait
	
	mov	coldata, col3
	rcall	setcol
	andi	temp, ~0x04
	out	PORTA, temp
	out	PORTB, temp2
	rcall	muxwait
	
	mov	coldata, col4
	rcall	setcol
	andi	temp2, ~0x08
	out	PORTA, temp
	out	PORTB, temp2
	rcall	muxwait
	
	mov	coldata, col5
	rcall	setcol
	andi	temp2, ~0x01
	out	PORTA, temp
	out	PORTB, temp2
	rcall	muxwait
	
	rjmp	loop
	
;--------------------

setcol:
	;convert column data (7 bits) to port outputs (2 bytes, PORTA and PORTB)
	ldi	temp, 0x34
	ldi	temp2, 0x09
	sbrs	coldata, 6
	ori	temp, 0x01
	sbrs	coldata, 5
	ori	temp, 0x02
	sbrs	coldata, 4
	ori	temp2, 0x02
	sbrs	coldata, 3
	ori	temp, 0x08
	sbrs	coldata, 2
	ori	temp2, 0x04
	sbrs	coldata, 1
	ori	temp, 0x80
	sbrs	coldata, 0
	ori	temp, 0x40
	ret

;--------------------

dec2bin:
	;convert 3-digit decimal number from RAM[X,X+2] to 1 byte binary
	ld	temp2, X+ ;first digit: multiply by 100 = (64 + 32 + 4)
	andi	temp2, 0x03
	lsl	temp2 ;temp2 = temp2 * 2
	mov	temp, temp2
	swap	temp ;temp = (temp2*2) * 16 = temp2 * 32
	lsl	temp2 ;temp2 = (temp2*2) * 2 = temp2 * 4
	add	temp, temp2
	swap	temp2 ;temp2 = ((temp2*2)*2) * 16 = temp2 * 64
	add	temp, temp2
	
	ld	temp2, X+ ;second digit: multiply by 10 = (8 + 2)
	andi	temp2, 0x0F
	lsl	temp2 ;temp2 = temp2 * 2
	add	temp, temp2
	lsl	temp2 ;temp2 = (temp2*2) * 4 = temp2 * 8
	lsl	temp2
	add	temp, temp2
	
	ld	temp2, X+ ;first digit
	andi	temp2, 0x0F
	add	temp, temp2
	
	ret
	
;--------------------

muxwait:
	;wait a short time
	ldi	temp, 2
	clr	temp2
	
muxwait_loop:
	dec	temp2
	brne	muxwait_loop
	dec	temp
	brne	muxwait_loop
	ret

;--------------------

waitbit:
	;wait 1/2 RS232 bit length (417 cycles - rcall - 2 cycles for saving the bit)
	ldi	temp2, 136
waitbit_loop:
	dec	temp2
	brne	waitbit_loop
	ret

;--------------------

eeprom_write:
	;write temp to EEPROM at ZL
	sbic	EECR, EEPE
	rjmp	eeprom_write
	out	EEARL, YL
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	ret

;--------------------

eeprom_read:
	;read temp from EEPROM at YL
	sbic	EECR, EEPE
	rjmp	eeprom_read
	out	EEARL, YL
	sbi	EECR, EERE
	in	temp, EEDR
	ret
